@@ -4,22 +4,12 @@
  *
  * Routines to support SELinux labels (security context)
  *
- * Copyright (c) 2010-2014, PostgreSQL Global Development Group
+ * Copyright (c) 2010-2016, PostgreSQL Global Development Group
  *
  * -------------------------------------------------------------------------
  */
 #include "postgres.h"
 
-#include <selinux/label.h>
-
-/*
- * <selinux/label.h> includes <stdbool.h>, which creates an incompatible
- * #define for bool.  Get rid of that so we can use our own typedef.
- * (We don't care if <stdbool.h> redefines "true"/"false"; those are close
- * enough.)
- */
-#undef bool
-
 #include "access/heapam.h"
 #include "access/htup_details.h"
 #include "access/genam.h"
@@ -47,6 +37,8 @@
 
 #include "sepgsql.h"
 
+#include <selinux/label.h>
+
 /*
  * Saved hook entries (if stacked)
  */
@@ -169,7 +161,7 @@ sepgsql_set_client_label(const char *new_label)
 /*
  * sepgsql_xact_callback
  *
- * A callback routine of transaction commit/abort/prepare.  Commmit or abort
+ * A callback routine of transaction commit/abort/prepare.  Commit or abort
  * changes in the client_label_pending list.
  */
 static void
@@ -540,7 +532,10 @@ sepgsql_object_relabel(const ObjectAddress *object, const char *seclabel)
 			break;
 
 		default:
-			elog(ERROR, "unsupported object type: %u", object->classId);
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+					 errmsg("sepgsql provider does not support labels on %s",
+							getObjectTypeDescription(object))));
 			break;
 	}
 }
